@@ -1,7 +1,6 @@
 # -*- coding: utf-8 -*-
 
 # print('__file__={0:<35} | __name__={1:<20} | __package__={2:<20}'.format(__file__, __name__, str(__package__)))
-from six import string_types
 from .core import *
 import pywinauto
 from win32api import GetCursorPos as win32api_GetCursorPos
@@ -13,8 +12,13 @@ from win32con import IDC_WAIT, MOUSEEVENTF_MOVE, MOUSEEVENTF_ABSOLUTE, MOUSEEVEN
     MOUSEEVENTF_MIDDLEDOWN, MOUSEEVENTF_MIDDLEUP, MOUSEEVENTF_RIGHTDOWN, MOUSEEVENTF_RIGHTUP, MOUSEEVENTF_WHEEL, WHEEL_DELTA
 import time
 from enum import Enum
+from typing import Any, Optional, Union, NewType
 
-# TODO special_char_array in core for recorder.py and player.py (check when too call escape & unescape)
+UI_Coordinates = NewType('UI_Coordinates', (float, float))
+UI_Element = Union[str, pywinauto.controls.uiawrapper.UIAWrapper, UI_Coordinates]
+
+
+# TODO special_char_array in core for recorder.py and player.py (check when to call escape & unescape)
 def unescape_special_char(string):
     for r in (("\\\\", "\\"), ("\\t", "\t"), ("\\n", "\n"), ("\\r", "\r"), ("\\v", "\v"), ("\\f", "\f"), ('\\"', '"')):
     #for r in (("\\", "\\\\"), ("\t", "\\t"), ("\n", "\\n"), ("\r", "\\r"), ("\v", "\\v"), ("\f", "\\f"), ('"', '\\"')):
@@ -38,8 +42,12 @@ element_path_old = ''
 w_rOLD = None
 
 
-def load_dictionary(filename):
-    
+def load_dictionary(filename: str) -> None:
+    """
+    Loads a dictionary
+
+    :param filename:filename of the dictionary
+    """
     with open(filename) as fp:
         for line in fp:
             words = line.split("\t")
@@ -51,13 +59,18 @@ def load_dictionary(filename):
             _dictionary[variable] = value
 
 
-def shortcut(variable):
-    return _dictionary[variable]
+def shortcut(str_shortcut: str) -> str:
+    """
+    Returns the element_path associated to th shortcut in the previously loaded dictionary
+
+    :param str_shortcut: shortcut
+    """
+    return _dictionary[str_shortcut]
 
 
 def wait_is_ready_try1(wrapper, timeout=120):
     """
-    Wait until element is ready (wait while greyed, not enabled, not visible, not ready, ...) :
+    Waits until element is ready (wait while greyed, not enabled, not visible, not ready, ...) :
     So far, I didn't find better than wait_cpu_usage_lower when greyed but must be enhanced
     """
     t0 = time.time()
@@ -111,7 +124,19 @@ class Region(object):
         Region.common_path = path_separator.join(self.list_path)
 
 
-def find(element_path=None, timeout=120):
+Window = Region
+
+
+def find(
+        element_path: UI_Element,
+        timeout: float = 120) -> UI_Element:
+    """
+    Finds an element
+
+    :param element_path: element path
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :return: Pywinauto wrapper of clicked element
+    """
     if not Region.click_desktop:
         Region.click_desktop = pywinauto.Desktop(backend='uia', allow_magic_lookup=False)
     if Region.common_path:
@@ -145,23 +170,26 @@ def find(element_path=None, timeout=120):
 
         _, _, y_x, _ = get_entry(entry_list[-1])
         if y_x is not None:
-            nb_y, nb_x, candidates = get_sorted_region(elements)
-            if is_int(y_x[0]):
-                unique_element = candidates[int(y_x[0])][int(y_x[1])]
+            if unique_element and int(y_x[0]) == int(y_x[1]) == 0:
+                return unique_element
             else:
-                ref_entry_list = get_entry_list(Region.common_path) + get_entry_list(y_x[0])
-                ref_unique_element, _ = find_element(
-                    Region.click_desktop, ref_entry_list, window_candidates=[], regex_title=Region.current.regex_title)
-                ref_r = ref_unique_element.rectangle()
-                r_y = 0
-                while r_y < nb_y:
-                    y_candidate = candidates[r_y][0].rectangle().mid_point()[1]
-                    if ref_r.top < y_candidate < ref_r.bottom:
-                        unique_element = candidates[r_y][y_x[1]]
-                        break
-                    r_y = r_y + 1
+                nb_y, nb_x, candidates = get_sorted_region(elements)
+                if is_int(y_x[0]):
+                    unique_element = candidates[int(y_x[0])][int(y_x[1])]
                 else:
-                    unique_element = None
+                    ref_entry_list = get_entry_list(Region.common_path) + get_entry_list(y_x[0])
+                    ref_unique_element, _ = find_element(
+                        Region.click_desktop, ref_entry_list, window_candidates=[], regex_title=Region.current.regex_title)
+                    ref_r = ref_unique_element.rectangle()
+                    r_y = 0
+                    while r_y < nb_y:
+                        y_candidate = candidates[r_y][0].rectangle().mid_point()[1]
+                        if ref_r.top < y_candidate < ref_r.bottom:
+                            unique_element = candidates[r_y][y_x[1]]
+                            break
+                        r_y = r_y + 1
+                    else:
+                        unique_element = None
         if unique_element is not None:
             break
         time.sleep(0.1)
@@ -170,13 +198,27 @@ def find(element_path=None, timeout=120):
     return unique_element
 
 
-def move(element_path, duration=0.5, mode=MoveMode.linear, timeout=120):
+def move(
+        element_path: UI_Element,
+        duration: Optional[float] = 0.5,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120) -> UI_Element:
+    """
+    Moves on element
+    
+    :param element_path: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :return: Pywinauto wrapper of clicked element
+    """
+    
     global unique_element_old
     global element_path_old
     global w_rOLD
-
+    
     x, y = win32api_GetCursorPos()
-    if isinstance(element_path, string_types):
+    if isinstance(element_path, str):
         element_path2 = element_path
         if Region.common_path:
             if Region.common_path != element_path[0:len(Region.common_path)]:
@@ -212,7 +254,7 @@ def move(element_path, duration=0.5, mode=MoveMode.linear, timeout=120):
             else:
                 dx, dy = 0, 0
             xd, yd = w_r.mid_point()
-            xd, yd = xd + round(dx/100.0*(w_r.width()-1), 0), round(yd + dy/100.0*(w_r.height()-1), 0)
+            xd, yd = xd + round(dx/100.0*(w_r.width()/2-1), 0), round(yd + dy/100.0*(w_r.height()/2-1), 0)
     elif issubclass(type(element_path), pywinauto.base_wrapper.BaseWrapper):
         unique_element = element_path
         element_path2 = get_wrapper_path(unique_element)
@@ -261,12 +303,30 @@ def move(element_path, duration=0.5, mode=MoveMode.linear, timeout=120):
     return unique_element
 
 
-def click(element_path, duration=0.5, mode=MoveMode.linear, button='left', timeout=120, wait_ready=True):
-    unique_element = move(element_path, duration=duration, mode=mode, timeout=timeout)
-    if wait_ready and isinstance(element_path, string_types):
-        wait_is_ready_try1(unique_element, timeout=timeout)
-    else:
-        unique_element = None
+def click(
+        element_path: Optional[UI_Element] = None,
+        duration: Optional[float] = 0.5,
+        mode: Enum = MoveMode.linear,
+        button: str = 'left',
+        timeout: float = 120,
+        wait_ready: bool = True) -> UI_Element:
+    """
+    Clicks on element
+    
+    :param element_path: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param button: mouse button: 'left','double_left', 'triple_left', 'right'
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :param wait_ready: if True waits until the element is ready
+    :return: Pywinauto wrapper of clicked element
+    """
+    if element_path:
+        unique_element = move(element_path, duration=duration, mode=mode, timeout=timeout)
+        if wait_ready and isinstance(element_path, str):
+            wait_is_ready_try1(unique_element, timeout=timeout)
+        else:
+            unique_element = None
     if button == 'left' or button == 'double_left' or button == 'triple_left':
         win32api_mouse_event(MOUSEEVENTF_LEFTDOWN, 0, 0)
         time.sleep(.01)
@@ -286,64 +346,191 @@ def click(element_path, duration=0.5, mode=MoveMode.linear, button='left', timeo
         time.sleep(.01)
         win32api_mouse_event(MOUSEEVENTF_RIGHTUP, 0, 0)
         time.sleep(.01)
-    return unique_element
+    if element_path:
+        return unique_element
+    else:
+        return None
 
 
-def left_click(element_path, duration=None, mode=MoveMode.linear, timeout=120, wait_ready=True):
+def left_click(
+        element_path: UI_Element,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120,
+        wait_ready: bool = True) -> UI_Element:
+    """
+    Left clicks on element
+    
+    :param element_path: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :param wait_ready: if True waits until the element is ready
+    :return: Pywinauto wrapper of clicked element
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
     return click(element_path, duration=duration, mode=mode, button='left', timeout=timeout, wait_ready=wait_ready)
 
 
-def right_click(element_path, duration=None, mode=MoveMode.linear, timeout=120, wait_ready=True):
+def right_click(
+        element_path: UI_Element,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120,
+        wait_ready: bool = True) -> UI_Element:
+    """
+    Right clicks on element
+    
+    :param element_path: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :param wait_ready: if True waits until the element is ready
+    :return: Pywinauto wrapper of clicked element
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
     return click(element_path, duration=duration, mode=mode, button='right', timeout=timeout, wait_ready=wait_ready)
 
 
-def double_left_click(element_path, duration=None, mode=MoveMode.linear, timeout=120, wait_ready=True):
+def double_left_click(
+        element_path: UI_Element,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120,
+        wait_ready: bool = True) -> UI_Element:
+    """
+    Double left clicks on element
+    
+    :param element_path: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :param wait_ready: if True waits until the element is ready
+    :return: Pywinauto wrapper of clicked element
+    """
+    
     if not duration:
         duration = PlayerSettings.mouse_move_duration
     return click(element_path, duration=duration, mode=mode, button='double_left', timeout=timeout, wait_ready=wait_ready)
 
 
-def triple_left_click(element_path, duration=None, mode=MoveMode.linear, timeout=120, wait_ready=True):
+def triple_left_click(
+        element_path: UI_Element,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120,
+        wait_ready: bool = True) -> UI_Element:
+    """
+    Triple left clicks on element
+    
+    :param element_path: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :param wait_ready: if True waits until the element is ready
+    :return: Pywinauto wrapper of clicked element
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
     return click(element_path, duration=duration, mode=mode, button='triple_left', timeout=timeout, wait_ready=wait_ready)
 
 
-def drag_and_drop(element_path1, element_path2, duration=None, mode=MoveMode.linear, timeout=120):
+def drag_and_drop(
+        element_path1: UI_Element,
+        element_path2: UI_Element,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120) -> UI_Element:
+    """
+    Drags and drop with left button pressed from element_path1 to element_path2.
+    
+    :param element_path1: element path
+    :param element_path2: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :return: Pywinauto wrapper with element_path2
+    """
+    
     if not duration:
         duration = PlayerSettings.mouse_move_duration
-    unique_element = move(element_path1, duration=duration, mode=mode, timeout=timeout)
+    move(element_path1, duration=duration, mode=mode, timeout=timeout)
     win32api_mouse_event(MOUSEEVENTF_LEFTDOWN, 0, 0)
-    move(element_path2, duration=duration, mode=mode)
+    unique_element = move(element_path2, duration=duration, timeout=timeout)
     win32api_mouse_event(MOUSEEVENTF_LEFTUP, 0, 0)
     return unique_element
 
 
-def middle_drag_and_drop(element_path1, element_path2, duration=None, mode=MoveMode.linear, timeout=120):
+def middle_drag_and_drop(
+        element_path1: UI_Element,
+        element_path2: UI_Element,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120) -> UI_Element:
+    """
+    Drags and drop with middle button pressed from element_path1 to element_path2.
+    
+    :param element_path1: element path
+    :param element_path2: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :return: Pywinauto wrapper with element_path2
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
-    unique_element = move(element_path1, duration=duration, mode=mode, timeout=timeout)
+    move(element_path1, duration=duration, mode=mode, timeout=timeout)
     win32api_mouse_event(MOUSEEVENTF_MIDDLEDOWN, 0, 0)
-    move(element_path2, duration=duration, mode=mode)
+    unique_element = move(element_path2, duration=duration, mode=mode, timeout=timeout)
     win32api_mouse_event(MOUSEEVENTF_MIDDLEUP, 0, 0)
     return unique_element
 
 
-def right_drag_and_drop(element_path1, element_path2, duration=None, mode=MoveMode.linear, timeout=120):
+def right_drag_and_drop(
+        element_path1: UI_Element,
+        element_path2: UI_Element,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120) -> UI_Element:
+    """
+    Drags and drop with right button pressed from element_path1 to element_path2.
+    
+    :param element_path1: element path
+    :param element_path2: element path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :return: Pywinauto wrapper with element_path2
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
-    unique_element = move(element_path1, duration=duration, mode=mode, timeout=timeout)
+    move(element_path1, duration=duration, mode=mode, timeout=timeout)
     win32api_mouse_event(MOUSEEVENTF_RIGHTDOWN, 0, 0)
-    move(element_path2, duration=duration, mode=mode)
+    unique_element = move(element_path2, duration=duration, mode=mode, timeout=timeout)
     win32api_mouse_event(MOUSEEVENTF_RIGHTUP, 0, 0)
     return unique_element
 
 
-def menu_click(element_path, menu_path, duration=None, mode=MoveMode.linear, menu_type='QT'):
+def menu_click(
+        element_path: UI_Element,
+        menu_path: str,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        menu_type: str = 'QT',
+        timeout: float = 120) -> UI_Element:
+    """
+    Clicks on menu item.
+    
+    :param element_path: element path
+    :param menu_path: menu path
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param menu_type: menu type ('QT', 'NPP')
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :return: Pywinauto wrapper of the clicked item
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
     menu_entry_list = menu_path.split(path_separator)
@@ -355,32 +542,33 @@ def menu_click(element_path, menu_path, duration=None, mode=MoveMode.linear, men
         element_path2 = element_path + path_separator
     else:
         element_path2 = ''
-    left_click(
-        element_path2 +
-        menu_entry_list[0] + type_separator + 'MenuBar' + path_separator +
-        menu_entry_list[1] + type_separator + 'MenuItem', duration=duration, mode=mode)
+    left_click(element_path2 +
+               menu_entry_list[0] + type_separator + 'MenuBar' + path_separator +
+               menu_entry_list[1] + type_separator + 'MenuItem', duration=duration, mode=mode, timeout=timeout)
     w = None
     if menu_type == 'QT':
         common_path_old = Region.common_path
         Region.common_path = ''
         for entry in menu_entry_list[2:]:
-            w = left_click(
-                type_separator + 'Menu' + path_separator +
-                entry + type_separator + 'MenuItem', duration=duration, mode=mode)
+            w = left_click(type_separator + 'Menu' + path_separator + entry + type_separator + 'MenuItem',
+                           duration=duration, mode=mode, timeout=timeout)
         Region.common_path = common_path_old
     else:
         for i, entry in enumerate(menu_entry_list[2:]):
-            w = left_click(
-                element_path +
-                menu_entry_list[i - 2] + type_separator + 'Menu' + path_separator +
-                unescape_special_char(entry) + type_separator + 'MenuItem', duration=duration, mode=mode)
+            w = left_click(element_path +
+                           menu_entry_list[i - 2] + type_separator + 'Menu' + path_separator +
+                           unescape_special_char(entry) + type_separator + 'MenuItem',
+                           duration=duration, mode=mode, timeout=timeout)
     return w
 
 
-Window = Region
-
-
-def mouse_wheel(steps, pause=0):
+def mouse_wheel(steps: int, pause: float = 0) -> None:
+    """
+    Turns the mouse wheel up or down.
+    
+    :param steps: number of wheel steps, if positive the mouse wheel is turned up else it is turned down
+    :param pause: pause in seconds between each wheel step
+    """
     if pause == 0:
         win32api_mouse_event(MOUSEEVENTF_WHEEL, 0, 0, WHEEL_DELTA * steps, 0)
     else:
@@ -393,15 +581,25 @@ def mouse_wheel(steps, pause=0):
 
 
 def send_keys(
-    str_keys,
-    pause=0.1,
-    with_spaces=True,
-    with_tabs=True,
-    with_newlines=True,
-    turn_off_numlock=True,
-    vk_packet=True):
+        str_keys: str,
+        pause: float = 0.1,
+        with_spaces: bool = True,
+        with_tabs: bool = True,
+        with_newlines: bool = True,
+        turn_off_numlock: bool = True,
+        vk_packet: bool = True) -> None:
     """
-    Parse the keys and type them
+    Parses the keys and type them
+    You can use any Unicode characters (on Windows) and some special keys.
+    See https://pywinauto.readthedocs.io/en/latest/code/pywinauto.keyboard.html
+    
+    :param str_keys: string representing the keys to be typed
+    :param pause: pause in seconds between each typed key
+    :param with_spaces: if False spaces are not taken into account
+    :param with_tabs: if False tabs are not taken into account
+    :param with_newlines: if False newlines are not taken into account
+    :param turn_off_numlock: if True numlock is turned off
+    :param vk_packet: For Windows only, pywinauto defaults to sending a virtual key packet (VK_PACKET) for textual input
     """
     for r in (('(', '{(}'),  (')', '{)}'), ('+', '{+}')):
         str_keys = str_keys.replace(*r)
@@ -416,24 +614,63 @@ def send_keys(
     )
 
 
-def set_combobox(element_path, text, duration=None, mode=MoveMode.linear, timeout=120):
+def set_combobox(
+        element_path: UI_Element,
+        value: str,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120) -> None:
+    """
+    Sets the value of a combobox.
+    
+    :param element_path: element path
+    :param value: value of the combobox
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
     left_click(element_path, duration=duration, mode=mode, timeout=timeout)
     time.sleep(0.9)
-    send_keys(text + "{ENTER}")
+    send_keys(value + "{ENTER}")
 
 
-def set_text(element_path, text, duration=None, mode=MoveMode.linear, timeout=120, pause=0.1):
+def set_text(
+        element_path: UI_Element,
+        value: str,
+        duration: Optional[float] = None,
+        mode: Enum = MoveMode.linear,
+        timeout: float = 120,
+        pause: float = 0.1) -> None:
+    """
+    Sets the value of a text field.
+    
+    :param element_path: element path
+    :param value: value of the combobox
+    :param duration: duration in seconds of the mouse move
+    :param mode: move mouse mode
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :param pause: pause in seconds between each typed key
+    """
     if not duration:
         duration = PlayerSettings.mouse_move_duration
     double_left_click(element_path, duration=duration, mode=mode, timeout=timeout)
     send_keys("{VK_CONTROL down}a{VK_CONTROL up}", pause=0)
     time.sleep(0.1)
-    send_keys(text + "{ENTER}", pause=pause)
+    send_keys(value + "{ENTER}", pause=pause)
 
 
-def exists(element_path, timeout=120):
+def exists(
+        element_path: UI_Element,
+        timeout: float = 120) -> UI_Element:
+    """
+    Tests if en UI_Element exists.
+    
+    :param element_path: element path
+    :param timeout: period of time in seconds that will be allowed to find the element
+    :return: Pywinauto wrapper of the found element or None
+    """
     try:
         wrapper = find(element_path, timeout=timeout)
         return wrapper
@@ -441,7 +678,18 @@ def exists(element_path, timeout=120):
         return None
 
 
-def select_file(element_path, full_path):
+def select_file(
+        element_path: UI_Element,
+        full_path: str,
+        force_slow_path_typing: bool = False) -> None:
+    """
+    Selects a file in an already opened file dialog.
+    
+    :param element_path: element path
+    :param full_path: the full path of the file to select
+    :param force_slow_path_typing: if True it will type the path even if the current path of the dialog box is the same
+    than the file to select
+    """
     import pyperclip
     import pathlib
     p = pathlib.Path(full_path)
@@ -449,8 +697,9 @@ def select_file(element_path, full_path):
     filename = p.name
     with Region(element_path, regex_title=True):
         left_click(find().descendants(title="All locations", control_type="SplitButton")[0])
-        send_keys("{VK_CONTROL down}""c""{VK_CONTROL up}")
-        if pathlib.Path(pyperclip.paste()) != folder:
+        if not force_slow_path_typing:
+            send_keys("{VK_CONTROL down}""c""{VK_CONTROL up}")
+        if force_slow_path_typing or pathlib.Path(pyperclip.paste()) != folder:
             send_keys(str(folder))
         send_keys("{ENTER}")
         double_left_click(u"File name:||ComboBox->File name:||Edit")
